@@ -7,6 +7,45 @@ from sklearn.preprocessing import MultiLabelBinarizer, OneHotEncoder, LabelEncod
 from sklearn.preprocessing import PolynomialFeatures
 
 
+class LabelEncoderExt(object):
+    def __init__(self):
+        """
+        It differs from LabelEncoder by handling new classes and providing a value for it [Unknown]
+        Unknown will be added in fit and transform will take care of new item. It gives unknown class id
+        """
+        self.label_encoder = LabelEncoder()
+        # self.classes_ = self.label_encoder.classes_
+
+    def fit(self, data_list):
+        """
+        This will fit the encoder for all the unique values and introduce unknown value
+        :param data_list: A list of string
+        :return: self
+        """
+        self.label_encoder = self.label_encoder.fit(list(data_list) + ['Unknown'])
+        self.classes_ = self.label_encoder.classes_
+
+        return self
+
+    def transform(self, data_list):
+        """
+        This will transform the data_list to id list where the new values get assigned to Unknown class
+        :param data_list:
+        :return:
+        """
+        new_data_list = list(data_list)
+        for unique_item in np.unique(data_list):
+            if unique_item not in self.label_encoder.classes_:
+                new_data_list = ['Unknown' if x==unique_item else x for x in new_data_list]
+
+        return self.label_encoder.transform(new_data_list)
+
+    def fit_transform(self, data_list):
+
+        self.fit(data_list)
+        return self.transform(data_list)
+
+
 def upload_blob(bucket_name, source_file_name, destination_blob_name):
     """Uploads a file to the bucket."""
     storage_client = storage.Client()
@@ -42,7 +81,7 @@ def create_equipment_type_multilabel_binarizer():
         "dumptruck",
         "conestoga",
         "hazmat",
-        ""]
+        u""]
 
     client = bigquery.Client()
 
@@ -104,7 +143,7 @@ def create_quantile_multilabel_binarizer():
 
 def create_one_hot_encoder_for_states():
     state_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    state_lbe = LabelEncoder()
+    state_lbe = LabelEncoderExt()
     client = bigquery.Client()
     states = []
 
@@ -170,6 +209,7 @@ def create_one_hot_encoder_for_states():
     for row in rows:
         states.append(row["state"])
 
+    states.append(u'')
     states = state_lbe.fit_transform(states)
     state_ohe.fit(states.reshape(-1, 1))
 
@@ -186,7 +226,7 @@ def create_one_hot_encoder_for_states():
 def create_one_hot_encoder_for_cities():
 
     city_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    city_lbe = LabelEncoder()
+    city_lbe = LabelEncoderExt()
     client = bigquery.Client()
     state_cities = []
 
@@ -504,6 +544,7 @@ def create_one_hot_encoder_for_cities():
     for row in rows:
         state_cities.append(row["city"])
 
+    state_cities.append(u'')
     state_cities = city_lbe.fit_transform(state_cities)
     city_ohe.fit(state_cities.reshape(-1, 1))
 
@@ -519,7 +560,7 @@ def create_one_hot_encoder_for_cities():
 
 def create_one_hot_encoder_load_source():
     source_ohe = OneHotEncoder(handle_unknown='ignore')
-    source_lbe = LabelEncoder()
+    source_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -536,6 +577,7 @@ def create_one_hot_encoder_load_source():
     for row in rows:
         sources.append(row["source"])
 
+    sources.append(u'')
     sources = source_lbe.fit_transform(sources)
     source_ohe.fit(sources.reshape(-1, 1))
 
@@ -550,7 +592,7 @@ def create_one_hot_encoder_load_source():
 def create_one_hot_encoder_email():
 
     email_ohe = OneHotEncoder(handle_unknown='ignore')
-    email_lbe = LabelEncoder()
+    email_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -567,6 +609,7 @@ def create_one_hot_encoder_email():
     for row in rows:
         emails.append(row["contact_email"])
 
+    emails.append(u'')
     emails = email_lbe.fit_transform(emails)
     email_ohe.fit(emails.reshape(-1, 1))
 
@@ -690,7 +733,7 @@ def create_polynomial_features(feature_name_list, input_df):
 
 def create_one_hot_encoder_for_user_zip_code():
     zip_code_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    zip_code_lbe = LabelEncoder()
+    zip_code_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -705,6 +748,7 @@ def create_one_hot_encoder_for_user_zip_code():
     for row in rows:
         zip_codes.append(row["zip_code"])
 
+    zip_codes.append(u'')
     zip_codes = zip_code_lbe.fit_transform(zip_codes)
     zip_code_ohe.fit(zip_codes.reshape(-1, 1))
 
@@ -718,7 +762,7 @@ def create_one_hot_encoder_for_user_zip_code():
 
 def create_one_hot_encoder_for_user_zip_code():
     zip_code_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    zip_code_lbe = LabelEncoder()
+    zip_code_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -733,6 +777,7 @@ def create_one_hot_encoder_for_user_zip_code():
     for row in rows:
         zip_codes.append(row["zip_code"])
 
+    zip_codes.append(u'')
     zip_codes = zip_code_lbe.fit_transform(zip_codes)
     zip_code_ohe.fit(zip_codes.reshape(-1, 1))
 
@@ -746,7 +791,7 @@ def create_one_hot_encoder_for_user_zip_code():
 
 def create_one_hot_encoder_for_user_dot():
     dot_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    dot_lbe = LabelEncoder()
+    dot_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -761,6 +806,7 @@ def create_one_hot_encoder_for_user_dot():
     for row in rows:
         dots.append(row["dot"])
 
+    dots.append(u'')
     dots = dot_lbe.fit_transform(dots)
     dot_ohe.fit(dots.reshape(-1, 1))
 
@@ -774,7 +820,7 @@ def create_one_hot_encoder_for_user_dot():
 
 def create_one_hot_encoder_for_user_email():
     email_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    email_lbe = LabelEncoder()
+    email_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -790,6 +836,7 @@ def create_one_hot_encoder_for_user_email():
     for row in rows:
         emails.append(row["email"])
 
+    emails.append(u'')
     emails = email_lbe.fit_transform(emails)
     email_ohe.fit(emails.reshape(-1, 1))
 
@@ -803,7 +850,7 @@ def create_one_hot_encoder_for_user_email():
 
 def create_one_hot_encoder_for_user_hos_cycle():
     hos_cycle_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    hos_cycle_lbe = LabelEncoder()
+    hos_cycle_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -819,6 +866,7 @@ def create_one_hot_encoder_for_user_hos_cycle():
     for row in rows:
         hos_cycles.append(row["hos_cycle"])
 
+    hos_cycles.append(u'')
     hos_cycles = hos_cycle_lbe.fit_transform(hos_cycles)
     hos_cycle_ohe.fit(hos_cycles.reshape(-1, 1))
 
@@ -832,7 +880,7 @@ def create_one_hot_encoder_for_user_hos_cycle():
 
 def create_one_hot_encoder_for_user_hos_home_timezone():
     hos_home_timezone_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    hos_home_timezone_lbe = LabelEncoder()
+    hos_home_timezone_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -848,6 +896,7 @@ def create_one_hot_encoder_for_user_hos_home_timezone():
     for row in rows:
         hos_home_timezones.append(row["hos_home_timezone"])
 
+    hos_home_timezones.append(u'')
     hos_home_timezones = hos_home_timezone_lbe.fit_transform(hos_home_timezones)
     hos_home_timezone_ohe.fit(hos_home_timezones.reshape(-1, 1))
 
@@ -861,7 +910,7 @@ def create_one_hot_encoder_for_user_hos_home_timezone():
 
 def create_one_hot_encoder_for_user_type():
     user_type_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    user_type_lbe = LabelEncoder()
+    user_type_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -877,6 +926,7 @@ def create_one_hot_encoder_for_user_type():
     for row in rows:
         user_types.append(row["user_type"])
 
+    user_types.append(u'')
     user_types = user_type_lbe.fit_transform(user_types)
     user_type_ohe.fit(user_types.reshape(-1, 1))
 
@@ -890,7 +940,7 @@ def create_one_hot_encoder_for_user_type():
 
 def create_one_hot_encoder_for_user_mc_number():
     mc_number_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    mc_number_lbe = LabelEncoder()
+    mc_number_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -906,6 +956,7 @@ def create_one_hot_encoder_for_user_mc_number():
     for row in rows:
         mc_numbers.append(row["mc_number"])
 
+    mc_numbers.append(u'')
     mc_numbers = mc_number_lbe.fit_transform(mc_numbers)
     mc_number_ohe.fit(mc_numbers.reshape(-1, 1))
 
@@ -919,7 +970,7 @@ def create_one_hot_encoder_for_user_mc_number():
 
 def create_one_hot_encoder_for_user_channel():
     channel_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    channel_lbe = LabelEncoder()
+    channel_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -935,6 +986,7 @@ def create_one_hot_encoder_for_user_channel():
     for row in rows:
         channels.append(row["channel"])
 
+    channels.append(u'')
     channels = channel_lbe.fit_transform(channels)
     channel_ohe.fit(channels.reshape(-1, 1))
 
@@ -948,7 +1000,7 @@ def create_one_hot_encoder_for_user_channel():
 
 def create_one_hot_encoder_for_safer_safety_rating():
     safer_result_safety_rating_ohe = OneHotEncoder(handle_unknown='ignore', sparse=True)
-    safer_result_safety_rating_lbe = LabelEncoder()
+    safer_result_safety_rating_lbe = LabelEncoderExt()
     client = bigquery.Client()
 
     query = """
@@ -964,6 +1016,7 @@ def create_one_hot_encoder_for_safer_safety_rating():
     for row in rows:
         ratings.append(row["safer_result_safety_rating"])
 
+    ratings.append(u'')
     safer_result_safety_ratings = safer_result_safety_rating_lbe.fit_transform(ratings)
     safer_result_safety_rating_ohe.fit(safer_result_safety_ratings.reshape(-1, 1))
 
@@ -973,3 +1026,5 @@ def create_one_hot_encoder_for_safer_safety_rating():
     pickle.dump(safer_result_safety_rating_lbe, open("user_safer_result_safety_rating_lbe.pkl", "wb"))
     upload_blob("user_transforms", "user_safer_result_safety_rating_lbe.pkl", "user_safer_result_safety_rating_lbe.pkl")
     return (safer_result_safety_rating_lbe, safer_result_safety_rating_ohe)
+
+
